@@ -30,9 +30,24 @@ IGNORE_SUMMARY_PRIMITIVES = {'d3m.primitives.data_transformation.construct_predi
                              'd3m.primitives.data_transformation.column_parser.Common'}
 
 
+def kernel_interrupt_handler(signal, frame):
+    logger.info('KeyboardInterrupt signal received, ending session...')
+    process = subprocess.Popen(['docker', 'stop', 'ta2_container'])
+    process.wait()
+    logger.info('Session ended!')
+
+    raise KeyboardInterrupt
+
+
 class Automl:
 
     def __init__(self, output_folder, ta2_id='NYU'):
+        """Create/instantiate an Automl object
+
+        :param output_folder: Path to the output directory
+        :param ta2_id: TA2 system name. It makes reference to the TA2 docker image. The provided TA2 systems are the
+            following: `NYU, CMU, SRI, TAMU`
+        """
         if ta2_id not in TA2_DOCKER_IMAGES:
             raise ValueError('Unknown "%s" TA2, you should choose among: [%s]' % (ta2_id, ', '.join(TA2_DOCKER_IMAGES)))
 
@@ -45,9 +60,36 @@ class Automl:
         self.leaderboard = None
         self.problem_config = None
 
-    def search_pipelines(self, dataset, time_bound, time_bound_run=10, target=None, metric=None, task_keywords=None,
+    def search_pipelines(self, dataset, time_bound, time_bound_run=5, target=None, metric=None, task_keywords=None,
+                         method='holdout', stratified=True, shuffle=True, folds=10, train_ratio=0.70, random_seed=0,
                          **kwargs):
+        """Perform the search of pipelines
+
+        :param dataset: Path to dataset. It supports D3M dataset, CSV file, OpenML, and Sklearn datasets
+        :param time_bound: Limit time in minutes to perform the search
+        :param time_bound_run: Limit time in minutes to score a pipeline
+        :param target: Column name of the potential target variable for a problem.
+        :param metric: The provided metrics are the following: `hammingLoss, accuracy, objectDetectionAP,
+            rocAucMicro, f1Macro, meanSquaredError, f1, jaccardSimilarityScore, normalizedMutualInformation, rocAuc,
+            f1Micro, hitsAtK, meanAbsoluteError, rocAucMacro, rSquared, recall, meanReciprocalRank, precision,
+            precisionAtTopK, rootMeanSquaredError`
+        :param task_keywords: A list of keywords that capture the nature of the machine learning task. The keywords
+            that can be combined to describe the task are the following: `tabular, nested, multiLabel, video,
+            linkPrediction, multivariate, graphMatching, forecasting, classification, graph, semiSupervised, text,
+            timeSeries, clustering, collaborativeFiltering, univariate, missingMetadata, remoteSensing, multiClass,
+            regression, multiGraph, lupi, relational, audio, grouped, objectDetection, vertexNomination,
+            communityDetection, geospatial, image, overlapping, nonOverlapping, speech, vertexClassification, binary`
+        :param method: Method to score the pipeline: `holdout, cross_validation`
+        :param stratified: Whether or not to split the data using a stratified strategy
+        :param shuffle: Whether or not to shuffle the data before splitting
+        :param folds: the seed used by the random number generator
+        :param train_ratio: Represent the proportion of the dataset to include in the train split
+        :param random_seed: The number seed used by the random generator
+        :param kwargs: Different arguments for problem's settings (e.g. pos_label for binary problems using F1)
+        :returns: List of pipelines
+        """
         suffix = 'TRAIN'
+        signal.signal(signal.SIGINT, kernel_interrupt_handler)
         if not is_d3m_format(dataset, suffix):
             self.problem_config = {'target_column': target, 'metric': metric, 'task_keywords': task_keywords,
                                    'optional': kwargs}
@@ -81,7 +123,8 @@ class Automl:
             duration = str(end_time - start_time)
             logger.info('Found pipeline id=%s, time=%s, scoring...' % (pipeline['id'], duration))
 
-            job = Thread(target=self.score_job, args=(pipeline, dataset_in_container, problem_path, self.pipelines))
+            job = Thread(target=self.score_in_search, args=(pipeline, dataset_in_container, problem_path, self.pipelines,
+                                                            method, stratified, shuffle, folds, train_ratio, random_seed))
             jobs.append(job)
             job.start()
 
@@ -104,53 +147,117 @@ class Automl:
         signal.alarm(0)
         return self.pipelines.values()
 
-    def train(self, solution_id):
+    def train(self, pipeline_id, expose_outputs=None):
+        """Train a model using an specific ML pipeline
+
+        :param pipeline_id: Pipeline id
+        :param expose_outputs: The output of the pipeline steps. If None, it doesn't expouse any output of the steps.
+            If str, should be 'all' to shows the output of each step in the pipeline, If list, it should contain the
+            ids of the steps, e.g. 'steps.2.produce'
+        :returns: A dictionary that contains the id and fitted model with/without the pipeline step outputs
+        """
+
         dataset_in_container = '/input/dataset/TRAIN/dataset_TRAIN/datasetDoc.json'
 
-        if solution_id not in self.pipelines:
-            raise ValueError('Pipeline id=%s does not exist' % solution_id)
+        if pipeline_id not in self.pipelines:
+            raise ValueError('Pipeline id=%s does not exist' % pipeline_id)
 
         logger.info('Training model...')
-        fitted_solution_id = self.ta3.do_train(solution_id, dataset_in_container)
-        fitted_solution = None  # TODO: Call to LoadFittedSolution, but TA2 could not have implemented it yet
-        model = {fitted_solution_id: fitted_solution}
+        if expose_outputs is None:
+            expose_outputs = []
+        elif isinstance(expose_outputs, str) and expose_outputs == 'all':
+            expose_outputs = ['outputs.0']
+            for index, step in enumerate(self.pipelines[pipeline_id]['json_representation']['steps']):
+                for id_output in step['outputs']:
+                    expose_outputs.append('steps.%d.%s' % (index, id_output['id']))
+
+        fitted_pipeline_id, pipeline_step_outputs = self.ta3.do_train(pipeline_id, dataset_in_container, expose_outputs)
+        fitted_pipeline = None  # TODO: Call to LoadFittedSolution, but TA2 could not have implemented it yet
+
+        for step_id, step_csv_uri in pipeline_step_outputs.items():
+            if not step_csv_uri.startswith('file://'):
+                logger.warning('Exposed step output "%s" cannot be read' % step_id)
+                continue
+            step_csv_uri = step_csv_uri.replace('file:///output/', '')
+            step_dataframe = pd.read_csv(join(self.output_folder, step_csv_uri))
+            pipeline_step_outputs[step_id] = step_dataframe
+
+        self.pipelines[pipeline_id]['fitted_id'] = fitted_pipeline_id
+        model = {'id': pipeline_id, 'fitted': fitted_pipeline}
         logger.info('Training finished!')
 
-        return model
+        if len(expose_outputs) == 0:
+            return model
 
-    def test(self, model, test_dataset):
+        return model, pipeline_step_outputs
+
+    def test(self, model, test_dataset, expose_outputs=None):
+        """Test a model
+
+        :param model: Dict that contains the id and fitted model
+        :param test_dataset: Path to dataset. It supports D3M dataset, and CSV file
+        :param expose_outputs: The output of the pipeline steps. If None, it doesn't expouse any output of the steps.
+            If str, should be 'all' to shows the output of each step in the pipeline, If list, it should contain the
+            ids of the steps, e.g. 'steps.2.produce'
+        :returns: A dataframe that contains the predictions with/without the pipeline step outputs
+        """
         suffix = 'TEST'
 
         if not is_d3m_format(test_dataset, suffix):
             convert_d3m_format(test_dataset, self.output_folder, self.problem_config, suffix)
 
         dataset_in_container = '/input/dataset/TEST/dataset_TEST/datasetDoc.json'
-        fitted_solution_id = list(model.keys())[0]
+        pipeline_id = model['id']
+        fitted_pipeline_id = self.pipelines[pipeline_id]['fitted_id']
         logger.info('Testing model...')
-        predictions_path_in_container = self.ta3.do_test(fitted_solution_id, dataset_in_container)
 
-        if not predictions_path_in_container.startswith('file://'):
-            raise FileNotFoundError('Exposed output "%s" from TA2 cannot be read', predictions_path_in_container)
+        if expose_outputs is None:
+            expose_outputs = []
+        elif isinstance(expose_outputs, str) and expose_outputs == 'all':
+            expose_outputs = ['outputs.0']
+            for index, step in enumerate(self.pipelines[pipeline_id]['json_representation']['steps']):
+                for id_output in step['outputs']:
+                    expose_outputs.append('steps.%d.%s' % (index, id_output['id']))
 
+        # Force to generate the predictions
+        if 'outputs.0' not in expose_outputs:
+            expose_outputs.append('outputs.0')
+
+        pipeline_step_outputs = self.ta3.do_test(fitted_pipeline_id, dataset_in_container, expose_outputs)
+
+        for step_id, step_csv_uri in pipeline_step_outputs.items():
+            if not step_csv_uri.startswith('file://'):
+                logger.warning('Exposed step output "%s" cannot be read' % step_id)
+                continue
+            step_csv_uri = step_csv_uri.replace('file:///output/', '')
+            step_dataframe = pd.read_csv(join(self.output_folder, step_csv_uri))
+            pipeline_step_outputs[step_id] = step_dataframe
+
+        predictions = pipeline_step_outputs['outputs.0']
         logger.info('Testing finished!')
-        predictions_path_in_container = predictions_path_in_container.replace('file:///output/', '')
-        predictions = pd.read_csv(join(self.output_folder, predictions_path_in_container))
 
-        return predictions
+        if len(expose_outputs) == 1:
+            return predictions
 
-    def score(self, solution_id, test_dataset):
+        return predictions, pipeline_step_outputs
+
+    def score(self, pipeline_id, test_dataset):
+        """Compute a proper score of the model
+
+        :param pipeline_id: Pipeline id
+        :param test_dataset: Path to dataset. It supports D3M dataset, and CSV file
+        :returns: A tuple holding metric name and score value
+        """
         suffix = 'SCORE'
 
-        if solution_id not in self.pipelines:
-            raise ValueError('Pipeline id=%s does not exist' % solution_id)
+        if pipeline_id not in self.pipelines:
+            raise ValueError('Pipeline id=%s does not exist' % pipeline_id)
 
         if not is_d3m_format(test_dataset, suffix):
             convert_d3m_format(test_dataset, self.output_folder, self.problem_config, suffix)
 
-        pipeline_id = self.pipelines[solution_id]['json_representation']['id']
-
         with open(join(self.output_folder, '%s.json' % pipeline_id), 'w') as fout:
-            json.dump(self.pipelines[solution_id]['json_representation'], fout)  # Save temporally the json pipeline
+            json.dump(self.pipelines[pipeline_id]['json_representation'], fout)  # Save temporally the json pipeline
 
         dataset_in_container = '/input/dataset/'
         dataset_train_path = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
@@ -180,7 +287,7 @@ class Automl:
         process.wait()
         result_path = join(self.output_folder, 'fit_score_%s.csv' % pipeline_id)
         if not exists(result_path):
-            raise FileNotFoundError('Pipeline id=%s could not be scored' % solution_id)
+            raise FileNotFoundError('Pipeline id=%s could not be scored' % pipeline_id)
 
         df = pd.read_csv(result_path)
         score = round(df['value'][0], 5)
@@ -189,6 +296,12 @@ class Automl:
         return metric, score
 
     def create_profiler_inputs(self, test_dataset=None):
+        """Create an proper input supported by PipelineProfiler based on the pipelines generated by a TA2 system
+
+        :param test_dataset: Path to dataset. If None it will use the search scores, otherwise will score the
+            pipelines over the passed dataset
+        :returns: List of pipelines in the PipelineProfiler input format
+        """
         profiler_inputs = []
         pipeline_ids = set()
 
@@ -240,6 +353,16 @@ class Automl:
 
         return profiler_inputs
 
+    def end_session(self):
+        """This safely ends session in D3M interface
+        """
+        logger.info('Ending session...')
+        if self.ta2 is not None:
+            process = subprocess.Popen(['docker', 'stop', 'ta2_container'])
+            process.wait()
+
+        logger.info('Session ended!')
+
     def start_ta2(self):
         logger.info('Initializing %s TA2...', self.ta2_id)
 
@@ -274,17 +397,11 @@ class Automl:
 
                 time.sleep(4)
 
-    def end_session(self):
-        logger.info('Ending session...')
-        if self.ta2 is not None:
-            process = subprocess.Popen(['docker', 'stop', 'ta2_container'])
-            process.wait()
-
-        logger.info('Session ended!')
-
-    def score_job(self, pipeline, dataset_in_container, problem_path, pipelines):
+    def score_in_search(self, pipeline, dataset_in_container, problem_path, pipelines, method, stratified, shuffle,
+                        folds, train_ratio, random_seed):
         try:
-            score_data = self.ta3.do_score(pipeline['id'], dataset_in_container, problem_path)
+            score_data = self.ta3.do_score(pipeline['id'], dataset_in_container, problem_path, method, stratified,
+                                           shuffle, folds, train_ratio, random_seed)
             logger.info('Scored pipeline id=%s, %s=%s' % (pipeline['id'], score_data['metric'], score_data['score']))
             pipeline['score'] = score_data['score']
             pipeline['normalized_score'] = score_data['normalized_score']
@@ -305,12 +422,22 @@ class Automl:
         return ', '.join(primitives_summary)
 
     @staticmethod
-    def add_new_ta2(name, docker_image):
-        TA2_DOCKER_IMAGES[name] = docker_image
-        logger.info('%s TA2 added!', name)
+    def add_new_ta2(ta2_id, docker_image_url):
+        """Add a new TA2 system that is not already defined in the D3M Interface. It can also be a different version of
+        a pre-existing TA2 (however it must be added with a different name)
+
+        :param ta2_id: A id to identify the new TA2
+        :param docker_image_url: The docker image url of the new TA2
+        """
+        TA2_DOCKER_IMAGES[ta2_id] = docker_image_url
+        logger.info('%s TA2 added!', ta2_id)
 
     @staticmethod
     def plot_summary_dataset(dataset_path):
+        """ Plot histograms of the dataset
+
+        :param dataset_path: Path to dataset.  It supports D3M dataset, and CSV file
+        """
         suffix = dataset_path.split('/')[-1]
 
         if is_d3m_format(dataset_path, suffix):
